/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others
Copyright (C) 2007-2008 Kristian Duske

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.

*/
// gl_sky.c
// eeeewww.  this was horrible stuff.  id sky code!  puke!

#include "quakedef.h"

#define skyvp \
	"!!ARBvp1.0\n" \
	"TEMP tc0;\n" \
	TRANSFORMVERTEX ("result.position", "vertex.attrib[0]") \
	"SUB tc0, vertex.attrib[1], program.env[0];\n" \
	"MUL tc0.z, tc0.z, 3.0;\n" \
	"MOV result.texcoord[0], tc0;\n" \
	"END\n"

#define skyfp \
	"!!ARBfp1.0\n" \
	"TEMP tc0, tc1, col1, col2;\n" \
	"DP3 tc0, fragment.texcoord[0], fragment.texcoord[0];\n" \
	"RSQ tc1.x, tc0.x;\n" \
	"MUL tc1.x, 378.0, tc1.x;\n" \
	"MUL tc0, fragment.texcoord[0], tc1.x;\n" \
	"ADD tc1, tc0, program.env[0].x;\n" \
	"MUL tc1, tc1, 0.0078125;\n" \
	"TEX col1, tc1, texture[0], 2D;\n" \
	"ADD tc1, tc0, program.env[0].y;\n" \
	"MUL tc1, tc1, 0.0078125;\n" \
	"TEX col2, tc1, texture[1], 2D;\n" \
	"TEMP fogmix;\n" \
	"LRP fogmix, col2.w, col2, col1;\n" \
	"LRP result.color, program.env[1].w, fogmix, program.env[1];\n" \
	"END\n"

GLuint arb_sky_vp = 0;
GLuint arb_sky_fp = 0;


#define skycubevp \
	"!!ARBvp1.0\n" \
	TRANSFORMVERTEX ("result.position", "vertex.attrib[0]") \
	"SUB result.texcoord[0], vertex.attrib[0], program.env[0];\n" \
	"END\n"

#define skycubefp \
	"!!ARBfp1.0\n" \
	"TEMP skycolor;\n" \
	"TEX skycolor, fragment.texcoord[0], texture[0], CUBE;\n" \
	"LRP result.color, program.env[1].w, skycolor, program.env[1];\n" \
	"END\n"

GLuint arb_skycube_vp = 0;
GLuint arb_skycube_fp = 0;


void Sky_CreateShaders (void)
{
	arb_sky_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, skyvp, false);
	arb_sky_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, skyfp, false);

	arb_skycube_vp = GL_CreateProgram (GL_VERTEX_PROGRAM_ARB, skycubevp, false);
	arb_skycube_fp = GL_CreateProgram (GL_FRAGMENT_PROGRAM_ARB, skycubefp, false);
}

float Fog_GetDensity (void);
float *Fog_GetColor (void);

extern	model_t	*loadmodel;

gltexture_t	*solidskytexture, *alphaskytexture;
char skybox_name[256] = {0};

qboolean sky_initialized = false;

cvar_t r_skyfog = {"r_skyfog", "0.5"};

int skytexorder[6] = {0, 2, 1, 3, 4, 5}; // for skybox


void GL_SkyBindSkybox (int layer)
{
}


// ==============================================================================
//
//  INIT
//
// ==============================================================================

/*
=============
Sky_LoadTexture

A sky texture is 256*128, with the left side being a masked overlay
==============
*/
void Sky_LoadTexture (texture_t *mt, byte *data)
{
	char		texturename[64];
	int			i, j;
	byte		*src;

	// these can't go into dynamic memory because the fucking texture manager needs them for reloading
	static byte	front_data[128 * 128];
	static byte	back_data[128 * 128];

	src = data;

	// extract back layer and upload
	for (i = 0; i < 128; i++)
		for (j = 0; j < 128; j++)
			back_data[(i * 128) + j] = src[i * 256 + j + 128];

	sprintf (texturename, "%s:%s_back", loadmodel->name, mt->name);
	solidskytexture = TM_LoadImage (loadmodel, texturename, 128, 128, SRC_INDEXED, back_data, "", (src_offset_t) back_data, TEXPREF_NONE);

	// extract front layer and upload
	for (i = 0; i < 128; i++)
	{
		for (j = 0; j < 128; j++)
		{
			front_data[(i * 128) + j] = src[i * 256 + j];

			// we need to store the original alpha colour so that we can properly set r_skyalpha
			if (front_data[(i * 128) + j] == 0)
			{
				front_data[(i * 128) + j] = 255;
			}
		}
	}

	sprintf (texturename, "%s:%s_front", loadmodel->name, mt->name);
	alphaskytexture = TM_LoadImage (loadmodel, texturename, 128, 128, SRC_INDEXED, front_data, "", (src_offset_t) front_data, TEXPREF_ALPHA);
}


// ==============================================================================
//
//  LOAD SKYBOX AS A CUBEMAP - DO NOT MESS WITH THIS; IT WORKS ON ATI NOW
//
// ==============================================================================

GLuint skyCubeTexture = 0;

void Sky_ClearSkybox (void)
{
	// purge old texture
	if (skyCubeTexture)
	{
		glDeleteTextures (1, &skyCubeTexture);
		skyCubeTexture = 0;
	}

	skybox_name[0] = 0;
}


void Sky_ReloadSkybox (void)
{
	char name[256];

	// copy out the skybox name and invalidate it to force a reload
	strcpy (name, skybox_name);
	Sky_ClearSkybox ();
	Sky_LoadCubeMap (name);
}


/*
==================
Sky_LoadCubeMap
==================
*/
char *suf[6] = {"rt", "bk", "lf", "ft", "up", "dn"};
int loadorder[6] = {0, 2, 1, 3, 4, 5};

GLenum cubefaces[6] =
{
	GL_TEXTURE_CUBE_MAP_POSITIVE_X,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_X,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Y,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Y,
	GL_TEXTURE_CUBE_MAP_POSITIVE_Z,
	GL_TEXTURE_CUBE_MAP_NEGATIVE_Z
};


void D3D_FlipTexels (unsigned int *texels, int width, int height)
{
	int x, y;

	for (x = 0; x < width; x++)
	{
		for (y = 0; y < (height / 2); y++)
		{
			int pos1 = y * width + x;
			int pos2 = (height - 1 - y) * width + x;

			unsigned int temp = texels[pos1];
			texels[pos1] = texels[pos2];
			texels[pos2] = temp;
		}
	}
}


void D3D_MirrorTexels (unsigned int *texels, int width, int height)
{
	int x, y;

	for (x = 0; x < (width / 2); x++)
	{
		for (y = 0; y < height; y++)
		{
			int pos1 = y * width + x;
			int pos2 = y * width + (width - 1 - x);

			unsigned int temp = texels[pos1];
			texels[pos1] = texels[pos2];
			texels[pos2] = temp;
		}
	}
}


void D3D_RotateTexelsInPlace (unsigned int *texels, int size)
{
	int i, j;

	for (i = 0; i < size; i++)
	{
		for (j = i; j < size; j++)
		{
			int pos1 = i * size + j;
			int pos2 = j * size + i;

			unsigned int temp = texels[pos1];

			texels[pos1] = texels[pos2];
			texels[pos2] = temp;
		}
	}
}


void GL_Resample32BitTexture (unsigned *in, int inwidth, int inheight, unsigned *out, int outwidth, int outheight);

void Sky_LoadCubeMap (char *name)
{
	int		i, mark, width[6], height[6];
	char	filename[MAX_OSPATH];
	byte	*data[6];
	qboolean nonefound = true;
	int largest;

	// so that I can call Sky_LoadCubeMap (NULL) to flush a texture.
	if (!name)
	{
		// purge old texture
		Sky_ClearSkybox ();
		return;
	}

	// no change
	if (strcmp (skybox_name, name) == 0) return;

	// purge old texture
	Sky_ClearSkybox ();

	// turn off skybox if sky is set to ""
	if (name[0] == 0)
	{
		skybox_name[0] = 0;
		return;
	}

	mark = Hunk_LowMark ();

	// skybox faces must all be square and the same dimension so track the largest
	largest = 0;

	// load textures
	for (i = 0; i < 6; i++)
	{
		sprintf (filename, "gfx/env/%s%s", name, suf[i]);
		data[i] = Image_LoadImage (filename, &width[i], &height[i]);

		if (data[i])
		{
			// skybox faces must all be square and the same dimension so track the largest
			if (width[i] > largest) largest = width[i];
			if (height[i] > largest) largest = height[i];
		}
		else width[i] = height[i] = 0;
	}

	// fixme - this could get a mite cleaner
	if (largest > 0)
	{
		// now let's see what we got
		byte *cubebuffer = NULL;

		glBindTexture (GL_TEXTURE_2D, 0);
		glDisable (GL_TEXTURE_2D);
		glEnable (GL_TEXTURE_CUBE_MAP);

		glGenTextures (1, &skyCubeTexture);
		glBindTexture (GL_TEXTURE_CUBE_MAP, skyCubeTexture);

		glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_S, GL_CLAMP_TO_EDGE);
		glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_T, GL_CLAMP_TO_EDGE);
		glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_WRAP_R, GL_CLAMP_TO_EDGE);

		glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
		glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

		// ATI strikes again - cubemaps should be loaded in a specific order
		for (i = 0; i < 6; i++)
		{
			if (!data[loadorder[i]])
			{
				if (!cubebuffer) cubebuffer = (byte *) Hunk_Alloc (largest * largest * 4);

				memset (cubebuffer, 0, largest * largest * 4);
				data[loadorder[i]] = cubebuffer;
				width[loadorder[i]] = largest;
				height[loadorder[i]] = largest;
			}

			if (width[loadorder[i]] != largest || height[loadorder[i]] != largest)
			{
				if (!cubebuffer) cubebuffer = (byte *) Hunk_Alloc (largest * largest * 4);

				// upsize to cube buffer and set back
				GL_Resample32BitTexture ((unsigned *) data[loadorder[i]], width[loadorder[i]], height[loadorder[i]], (unsigned *) cubebuffer, largest, largest);

				data[loadorder[i]] = cubebuffer;
				width[loadorder[i]] = largest;
				height[loadorder[i]] = largest;
			}

			switch (loadorder[i])
			{
			case 0:
				D3D_RotateTexelsInPlace ((unsigned int *) data[0], width[0]);
				break;

			case 1:
				D3D_FlipTexels ((unsigned int *) data[1], width[1], height[1]);
				break;

			case 2:
				D3D_RotateTexelsInPlace ((unsigned int *) data[2], width[2]);
				D3D_MirrorTexels ((unsigned int *) data[2], width[2], height[2]);
				D3D_FlipTexels ((unsigned int *) data[2], width[2], height[2]);
				break;

			case 3:
				D3D_MirrorTexels ((unsigned int *) data[3], width[3], height[3]);
				break;

			case 4:
				D3D_RotateTexelsInPlace ((unsigned int *) data[4], width[4]);
				break;

			case 5:
				D3D_RotateTexelsInPlace ((unsigned int *) data[5], width[5]);
				break;
			}

			// standard face
			glTexImage2D (cubefaces[i], 0, GL_RGBA8, largest, largest, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, data[loadorder[i]]);

			nonefound = false;
		}

		glBindTexture (GL_TEXTURE_CUBE_MAP, 0);
		glDisable (GL_TEXTURE_CUBE_MAP);
		glEnable (GL_TEXTURE_2D);
		GL_BindTexture (GL_TEXTURE0, NULL);
	}

	Hunk_FreeToLowMark (mark);

	if (nonefound) // go back to scrolling sky if skybox is totally missing
	{
		Sky_ClearSkybox ();
		Con_Printf ("Couldn't load %s\n", name);
		return;
	}

	strcpy (skybox_name, name);
	Con_DPrintf ("loaded skybox %s OK\n", name);
}


/*
=================
Sky_NewMap
=================
*/
char *sbkeys[] =
{
	"sky",
	"skyname",
	"q1sky",
	"skybox",
	NULL
};

void Sky_NewMap (void)
{
	char	key[128], value[4096];
	char	*data;
	int		i;

	// initially no sky
	Sky_ClearSkybox ();

	// read worldspawn (this is so ugly, and shouldn't it be done on the server?)
	data = cl.worldmodel->entities;

	if (!data)
		return; // FIXME: how could this possibly ever happen? -- if there's no
				// worldspawn then the sever wouldn't send the loadmap message to the client

	data = COM_Parse (data);

	if (!data) // should never happen
		return; // error

	if (com_token[0] != '{') // should never happen
		return; // error

	while (1)
	{
		data = COM_Parse (data);

		if (!data)
			return; // error

		if (com_token[0] == '}')
			break; // end of worldspawn

		if (com_token[0] == '_')
			strcpy (key, com_token + 1);
		else strcpy (key, com_token);

		while (key[strlen (key) - 1] == ' ') // remove trailing spaces
			key[strlen (key) - 1] = 0;

		data = COM_Parse (data);

		if (!data)
			return; // error

		strcpy (value, com_token);

		// better key testing
		for (i = 0; ; i++)
		{
			// no more
			if (!sbkeys[i]) break;

			if (!strcmp (sbkeys[i], key))
			{
				skybox_name[0] = 0;
				Sky_LoadCubeMap (value);
			}
		}
	}
}

/*
=================
Sky_SkyCommand_f
=================
*/
void Sky_SkyCommand_f (void)
{
	switch (Cmd_Argc ())
	{
	case 1:
		// give the correct command
		Con_Printf ("\"%s\" is \"%s\"\n", Cmd_Argv (0), skybox_name);
		break;

	case 2:
		Sky_LoadCubeMap (Cmd_Argv (1));
		break;

	default:
		// give the correct command
		Con_Printf ("usage: %s <skyname>\n", Cmd_Argv (0));
	}
}

/*
=============
Sky_Init
=============
*/
void Sky_Init (void)
{
	Cvar_RegisterVariable (&r_skyfog, NULL);

	// mh - every other engine supports loadsky so lets have some consistency around here please
	Cmd_AddCommand ("sky", Sky_SkyCommand_f);
	Cmd_AddCommand ("loadsky", Sky_SkyCommand_f);
}


// ==============================================================================
//
//  RENDERER
//
// ==============================================================================

void Sky_SetFog (void)
{
	float params[4] = {0, 0, 0, 0};

	if (Fog_GetDensity () > 0 && r_skyfog.value > 0)
	{
		float *c = Fog_GetColor ();

		// every time the illogical order of this has caught me out... grrr...
		VectorCopy (c, params);
		params[3] = 1.0f - r_skyfog.value;
	}
	else params[0] = params[1] = params[2] = params[3] = 1.0f;

	qglProgramEnvParameter4fvARB (GL_FRAGMENT_PROGRAM_ARB, 1, params);
}


void R_DrawSkySurfaces (msurface_t *chain, entity_t *ent)
{
	msurface_t *surf;
	float params[4] = {0, 0, 0, 0};

	// common
	Fog_DisableGlobalFog ();

	if (skybox_name[0])
	{
		// yayyy shaders!  even ricky's ATI runs this now.  bye bye mess.
		// (fixme - really MUST mod texmgr to support cubemaps.  God but it's fragile.  Might even just do a
		// glGetTexImage instead of the stuff in there.)
		glBindTexture (GL_TEXTURE_CUBE_MAP, skyCubeTexture);
	}
	else
	{
		GL_BindTexture (GL_TEXTURE0, solidskytexture);
		GL_BindTexture (GL_TEXTURE1, alphaskytexture);

		// fixme - make scroll speeds configurable???
		params[0] = cl.time * 8.0f;
		params[0] -= (int) params[0] & ~127;

		params[1] = cl.time * 16.0f;
		params[1] -= (int) params[1] & ~127;

		qglProgramEnvParameter4fvARB (GL_FRAGMENT_PROGRAM_ARB, 0, params);
	}

	params[0] = r_origin[0];
	params[1] = r_origin[1];
	params[2] = r_origin[2];

	qglProgramEnvParameter4fvARB (GL_VERTEX_PROGRAM_ARB, 0, params);

	Sky_SetFog ();

	for (surf = chain; surf; surf = surf->texturechain)
	{
		if (skybox_name[0])
		{
			surf->vertexprogram = arb_skycube_vp;
			surf->fragmentprogram = arb_skycube_fp;
		}
		else
		{
			surf->vertexprogram = arb_sky_vp;
			surf->fragmentprogram = arb_sky_fp;
		}

		R_PrecheckSurface (surf, ent);
		R_BatchSurface (surf);
	}

	R_FlushSurfaces ();
	Fog_EnableGlobalFog ();
}


