/*
Copyright (C) 1996-2001 Id Software, Inc.
Copyright (C) 2002-2009 John Fitzgibbons and others

This program is free software; you can redistribute it and/or
modify it under the terms of the GNU General Public License
as published by the Free Software Foundation; either version 2
of the License, or (at your option) any later version.

This program is distributed in the hope that it will be useful,
but WITHOUT ANY WARRANTY; without even the implied warranty of
MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.

See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License
along with this program; if not, write to the Free Software
Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
*/

#ifndef _GL_TEXMAN_H
#define _GL_TEXMAN_H

// gl_texmgr.h -- fitzquake's texture manager. manages opengl texture images

#define TEXPREF_NONE			0x00000000
#define TEXPREF_MIPMAP			0x00000001	// generate mipmaps
#define TEXPREF_LINEAR			0x00000002	// force linear
#define TEXPREF_NEAREST			0x00000004	// force nearest
#define TEXPREF_ALPHA			0x00000008	// allow alpha
#define TEXPREF_PAD				0x00000010	// allow padding
#define TEXPREF_PERSIST			0x00000020	// never free
#define TEXPREF_OVERWRITE		0x00000040	// overwrite existing same-name texture
#define TEXPREF_NOPICMIP		0x00000080	// always load full-sized
#define TEXPREF_FULLBRIGHT		0x00000100	// use fullbright mask palette
#define TEXPREF_NOBRIGHT		0x00000200	// use nobright mask palette
#define TEXPREF_CONCHARS		0x00000400	// use conchars palette
#define TEXPREF_FLOODFILL		0x00000800	// use conchars palette
#define TEXPREF_CLAMPTOEDGE		0x00001000	// use edge clamping
#define TEXPREF_CLAMP			0x00002000	// use standard clamping

enum srcformat {SRC_INDEXED, SRC_LIGHTMAP, SRC_RGBA};

typedef size_t src_offset_t; // technically need uintptr_t, but size_t is OK virtually everywhere.


typedef struct gltexture_s
{
	// managed by texture manager
	unsigned int		texnum;
	struct gltexture_s	*next;
	model_t				*owner;
	// managed by image loading
	char				name[64];
	unsigned int		width; // size of image as it exists in opengl
	unsigned int		height; // size of image as it exists in opengl
	unsigned int		flags;
	char				source_file[MAX_QPATH]; // relative filepath to data source, or "" if source is in memory
	src_offset_t		source_offset; // byte offset into file, or memory address
	enum srcformat		source_format; // format of pixel data (indexed, lightmap, or rgba)
	unsigned int		source_width; // size of image in source data
	unsigned int		source_height; // size of image in source data
	unsigned short		source_crc; // generated by source data before modifications
	char				shirt; // 0-13 shirt color, or -1 if never colormapped
	char				pants; // 0-13 pants color, or -1 if never colormapped
	// used for rendering
	int					visframe; // matches r_framecount if texture was bound this frame
} gltexture_t;

extern gltexture_t *notexture;
extern gltexture_t *nulltexture;

extern unsigned int d_8to24table[256];
extern unsigned int d_8to24table_rgba[256];
extern unsigned int d_8to24table_fbright[256];
extern unsigned int d_8to24table_nobright[256];
extern unsigned int d_8to24table_conchars[256];
extern unsigned int d_8to24table_shirt[256];
extern unsigned int d_8to24table_pants[256];

// TEXTURE MANAGER

float TM_FrameUsage (void);
gltexture_t *TM_FindTexture (model_t *owner, char *name);
gltexture_t *TM_NewTexture (void);
void TM_FreeTexture (gltexture_t *kill);
void TM_FreeTextures (int flags, int mask);
void TM_FreeTexturesForOwner (model_t *owner);
void TM_NewGame (void);
void TM_Init (void);

// IMAGE LOADING
gltexture_t *TM_LoadImage (model_t *owner, char *name, int width, int height, enum srcformat format,
							   byte *data, char *source_file, src_offset_t source_offset, unsigned flags);
void TM_ReloadImage (gltexture_t *glt, int shirt, int pants);
void TM_ReloadImages (void);

int TM_Pad (int s);
int TM_SafeTextureSize (int s);
int TM_PadConditional (int s);

// TEXTURE BINDING & TEXTURE UNIT SWITCHING

extern qboolean gl_mtexable;

void GL_BindTexture (GLenum tmu, gltexture_t *texture);
void GL_UnbindTextures (void);

#endif	/* _GL_TEXMAN_H */

