
#define WINDOW_WIDTH	800
#define WINDOW_HEIGHT	600

#include "SDL.h"
#include "SDL_opengl.h"

#pragma comment (lib, "SDL.lib")
#pragma comment (lib, "SDLmain.lib")
#pragma comment (lib, "opengl32.lib")
#pragma comment (lib, "glu32.lib")

GLuint CubeTexture = 0;
GLuint MainTexture = 0;

void CheckGLError (char *func)
{
	if (glGetError () != GL_NO_ERROR)
	{
		printf ("GL Error on %s\n", func);
	}
}


void CubeVertex (float x, float y, float z)
{
	glTexCoord3f (x, y, z);
	CheckGLError ("glTexCoord3f (x, y, z);");

	glVertex3f (x, y, z);
	CheckGLError ("glVertex3f (x, y, z);");
}


void RenderOpenGL (void)
{
	Uint32 ticks = SDL_GetTicks ();
	float ang = (float) ticks;

	ang /= 25.0f;

	while (ang > 360) ang -= 360;

	glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
	CheckGLError ("glClear (GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);");

	glViewport (0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);
	CheckGLError ("glViewport (0, 0, WINDOW_WIDTH, WINDOW_HEIGHT);");

	glMatrixMode (GL_PROJECTION);
	CheckGLError ("glMatrixMode (GL_PROJECTION);");
	glLoadIdentity ();
	CheckGLError ("glLoadIdentity ();");
	gluPerspective (45, (float) WINDOW_WIDTH / (float) WINDOW_HEIGHT, 10, 400);
	CheckGLError ("gluPerspective (45, (float) WINDOW_WIDTH / (float) WINDOW_HEIGHT, 10, 400);");

	glMatrixMode (GL_MODELVIEW);
	CheckGLError ("glMatrixMode (GL_MODELVIEW);");
	glLoadIdentity ();
	CheckGLError ("glLoadIdentity ();");
	glTranslatef (0, 0, -13);
	CheckGLError ("glTranslatef (0, 0, -13);");
	glRotatef (ang, 1, 1, 1);
	CheckGLError ("glRotatef (ang, 1, 1, 1);");

	glEnable (GL_DEPTH_TEST);
	CheckGLError ("glEnable (GL_DEPTH_TEST);");

	glBindTexture (GL_TEXTURE_2D, 0);
	CheckGLError ("glBindTexture (GL_TEXTURE_2D, 0);");

	glDisable (GL_TEXTURE_2D);
	CheckGLError ("glDisable (GL_TEXTURE_2D);");
	glEnable (GL_TEXTURE_CUBE_MAP);
	CheckGLError ("glEnable (GL_TEXTURE_CUBE_MAP);");

	glBindTexture (GL_TEXTURE_CUBE_MAP, CubeTexture);
	CheckGLError ("glBindTexture (GL_TEXTURE_CUBE_MAP, CubeTexture);");

	glBegin (GL_QUADS);
	CheckGLError ("glBegin (GL_QUADS);");

	CubeVertex (1.0f, 1.0f, -1.0f);
	CubeVertex (-1.0f, 1.0f, -1.0f);
	CubeVertex (-1.0f, 1.0f, 1.0f);
	CubeVertex (1.0f, 1.0f, 1.0f);

	CubeVertex (1.0f, -1.0f, 1.0f);
	CubeVertex (-1.0f, -1.0f, 1.0f);
	CubeVertex (-1.0f, -1.0f, -1.0f);
	CubeVertex (1.0f, -1.0f, -1.0f);

	CubeVertex (1.0f, 1.0f, 1.0f);
	CubeVertex (-1.0f, 1.0f, 1.0f);
	CubeVertex (-1.0f, -1.0f, 1.0f);
	CubeVertex (1.0f, -1.0f, 1.0f);

	CubeVertex (1.0f, -1.0f, -1.0f);
	CubeVertex (-1.0f, -1.0f, -1.0f);
	CubeVertex (-1.0f, 1.0f, -1.0f);
	CubeVertex (1.0f, 1.0f, -1.0f);

	CubeVertex (-1.0f, 1.0f, 1.0f);
	CubeVertex (-1.0f, 1.0f, -1.0f);
	CubeVertex (-1.0f, -1.0f, -1.0f);
	CubeVertex (-1.0f, -1.0f, 1.0f);

	CubeVertex (1.0f, 1.0f, -1.0f);
	CubeVertex (1.0f, 1.0f, 1.0f);
	CubeVertex (1.0f, -1.0f, 1.0f);
	CubeVertex (1.0f, -1.0f, -1.0f);

	glEnd ();
	glGetError ();	// glEnd throws an error - WTF?

	glBindTexture (GL_TEXTURE_CUBE_MAP, 0);
	CheckGLError ("glBindTexture (GL_TEXTURE_CUBE_MAP, 0);");

	glDisable (GL_TEXTURE_CUBE_MAP);
	CheckGLError ("glDisable (GL_TEXTURE_CUBE_MAP);");
	glEnable (GL_TEXTURE_2D);
	CheckGLError ("glEnable (GL_TEXTURE_2D);");

	glLoadIdentity ();
	glTranslatef (0, 0, -20);
	glRotatef (-ang, 1, 1, 1);

	glBindTexture (GL_TEXTURE_2D, MainTexture);
	CheckGLError ("glBindTexture (GL_TEXTURE_2D, MainTexture);");

	glBegin (GL_QUADS);

	CubeVertex (4.0f, -4.0f, 0);
	CubeVertex (-4.0f, -4.0f, 0);
	CubeVertex (-4.0f, 4.0f, 0);
	CubeVertex (4.0f, 4.0f, 0);

	glEnd ();
	glGetError ();	// glEnd throws an error - WTF?
}


void MakeCubeFace (int red, int green, int blue, int mult, GLenum face)
{
	unsigned char *data = (unsigned char *) malloc (32 * 32 * 4);
	unsigned char *dataptr = data;

	for (int i = 0; i < 32 * 32; i++, dataptr += 4)
	{
		dataptr[0] = (blue + mult * i) & 255;
		dataptr[1] = (green + mult * i) & 255;
		dataptr[2] = (red + mult * i) & 255;
		dataptr[3] = 255;
	}

	glTexImage2D (face, 0, GL_RGBA8, 32, 32, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, data);
	CheckGLError ("glTexImage2D (face, 0, GL_RGBA8, 32, 32, 0, GL_BGRA, GL_UNSIGNED_INT_8_8_8_8_REV, data);");

	free (data);
}


void InitGLObjects (void)
{
	glBindTexture (GL_TEXTURE_2D, 0);
	CheckGLError ("glBindTexture (GL_TEXTURE_2D, 0);");

	glDisable (GL_TEXTURE_2D);
	CheckGLError ("glDisable (GL_TEXTURE_2D);");
	glEnable (GL_TEXTURE_CUBE_MAP);
	CheckGLError ("glEnable (GL_TEXTURE_CUBE_MAP);");

	glGenTextures (1, &CubeTexture);
	CheckGLError ("glGenTextures (1, &CubeTexture);");
	glBindTexture (GL_TEXTURE_CUBE_MAP, CubeTexture);
	CheckGLError ("glBindTexture (GL_TEXTURE_CUBE_MAP, CubeTexture);");

	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	CheckGLError ("glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MAG_FILTER, GL_LINEAR);");
	glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);
	CheckGLError ("glTexParameteri (GL_TEXTURE_CUBE_MAP, GL_TEXTURE_MIN_FILTER, GL_LINEAR);");

	MakeCubeFace (255, 0, 0, 0, GL_TEXTURE_CUBE_MAP_POSITIVE_X);
	MakeCubeFace (0, 255, 0, 0, GL_TEXTURE_CUBE_MAP_NEGATIVE_X);
	MakeCubeFace (0, 0, 255, 0, GL_TEXTURE_CUBE_MAP_POSITIVE_Y);
	MakeCubeFace (255, 255, 0, 0, GL_TEXTURE_CUBE_MAP_NEGATIVE_Y);
	MakeCubeFace (255, 0, 255, 0, GL_TEXTURE_CUBE_MAP_POSITIVE_Z);
	MakeCubeFace (0, 255, 255, 0, GL_TEXTURE_CUBE_MAP_NEGATIVE_Z);

	glBindTexture (GL_TEXTURE_CUBE_MAP, 0);
	CheckGLError ("glBindTexture (GL_TEXTURE_CUBE_MAP, 0);");
	glDisable (GL_TEXTURE_CUBE_MAP);
	CheckGLError ("glDisable (GL_TEXTURE_CUBE_MAP);");
	glEnable (GL_TEXTURE_2D);
	CheckGLError ("glEnable (GL_TEXTURE_2D);");

	glGenTextures (1, &MainTexture);
	glBindTexture (GL_TEXTURE_2D, MainTexture);

	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_LINEAR);
	glTexParameteri (GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_LINEAR);

	MakeCubeFace (255, 128, 0, 1, GL_TEXTURE_2D);

	glGetError ();
}


int main (int argc, char *argv[])
{
	if (SDL_Init (SDL_INIT_VIDEO | SDL_INIT_NOPARACHUTE) != 0)
	{
		printf ("Unable to initialize SDL: %s\n", SDL_GetError ());
		return 1;
	}

	SDL_GL_SetAttribute (SDL_GL_DOUBLEBUFFER, 1);
	SDL_GL_SetAttribute (SDL_GL_DEPTH_SIZE, 24);
	SDL_Surface *screen = SDL_SetVideoMode (WINDOW_WIDTH, WINDOW_HEIGHT, 32, SDL_OPENGL);

	SDL_WM_SetCaption ("Press ESC to exit", "");
	glClearColor (0, 0, 0, 1);

	int done = 0;
	SDL_Event evt;

	InitGLObjects ();

	while (!done)
	{
		while (!done && SDL_PollEvent (&evt))
		{
			if (evt.type == SDL_QUIT)
			{
				done = 1;
				break;
			}
			else if (evt.type == SDL_KEYUP)
			{
				if (evt.key.keysym.sym == SDLK_ESCAPE)
				{
					done = 1;
					break;
				}
			}
		}

		// run the screen update here
		RenderOpenGL ();
		SDL_GL_SwapBuffers ();
	}

	return 0;
}

